;;; multiple-value-bind.lisp
;;;
;;; Copyright (C) 2004 Peter Graves
;;; $Id$
;;;
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;;;
;;; As a special exception, the copyright holders of this library give you
;;; permission to link this library with independent modules to produce an
;;; executable, regardless of the license terms of these independent
;;; modules, and to copy and distribute the resulting executable under
;;; terms of your choice, provided that you also meet, for each linked
;;; independent module, the terms and conditions of the license of that
;;; module.  An independent module is a module which is not derived from
;;; or based on this library.  If you modify this library, you may extend
;;; this exception to your version of the library, but you are not
;;; obligated to do so.  If you do not wish to do so, delete this
;;; exception statement from your version.

;;; Adapted from CMUCL.

(in-package "SYSTEM")

;; The traditional implementation of M-V-B in terms of M-V-C. ABCL implements
;; M-V-B as a special form in the interpreter, and ABCL's compiler handles it
;; specifically too, so this code is only here to support code walkers and the
;; like, as required by ANSI.
(defmacro multiple-value-bind (varlist value-form &body body)
  (unless (and (listp varlist) (every #'symbolp varlist))
    (error 'program-error
           :format-control "Variable list is not a list of symbols: ~S."
           :format-arguments (list varlist)))
  (if (= (length varlist) 1)
      `(let ((,(car varlist) ,value-form))
         ,@body)
      (let ((ignore (gensym)))
        `(multiple-value-call #'(lambda (&optional ,@(mapcar #'list varlist) &rest ,ignore)
                                 (declare (ignore ,ignore))
                                 ,@body)
                              ,value-form))))
