;;; check-type.lisp
;;;
;;; Copyright (C) 2003-2005 Peter Graves
;;; $Id$
;;;
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;;;
;;; As a special exception, the copyright holders of this library give you
;;; permission to link this library with independent modules to produce an
;;; executable, regardless of the license terms of these independent
;;; modules, and to copy and distribute the resulting executable under
;;; terms of your choice, provided that you also meet, for each linked
;;; independent module, the terms and conditions of the license of that
;;; module.  An independent module is a module which is not derived from
;;; or based on this library.  If you modify this library, you may extend
;;; this exception to your version of the library, but you are not
;;; obligated to do so.  If you do not wish to do so, delete this
;;; exception statement from your version.

;;; Adapted from CMUCL.

(in-package #:system)

(defmacro check-type (place type &optional type-string)
  (let ((place-value (gensym)))
    `(loop
       (let ((,place-value ,place))
         (when (typep ,place-value ',type)
           (return nil))
         (setf ,place
               (check-type-error ',place ,place-value ',type ,type-string))))))

(defun check-type-error (place place-value type type-string)
  (let ((cond (if type-string
                  (make-condition 'simple-type-error
                                  :datum place-value :expected-type type
                                  :format-control
                                  "The value of ~S is ~S, which is not ~A."
                                  :format-arguments
                                  (list place place-value type-string))
                  (make-condition 'simple-type-error
                                  :datum place-value :expected-type type
                                  :format-control
                                  "The value of ~S is ~S, which is not of type ~S."
                                  :format-arguments
                                  (list place place-value type)))))
    (restart-case (error cond)
      (store-value (value)
        :report (lambda (stream)
                  (format stream "Supply a new value for ~S."
                          place))
        :interactive read-evaluated-form
        value))))
